/*
FRODO: a FRamework for Open/Distributed Optimization
Copyright (C) 2008, 2009, 2010, 2011  Thomas Leaute, Brammert Ottens & Radoslaw Szymanek

FRODO is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FRODO is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


How to contact the authors: 
<http://lia.epfl.ch/>

EPFL / IC / IIF / LIA
Batiment IN 
Station 14 
CH - 1015 Lausanne 
Switzerland
*/

/** StochDCOP algorithms that maximize worst-case rather than expected utility */
package ch.epfl.lia.frodo.algorithms.dpop.stochastic.robust;

import java.util.ArrayList;
import java.util.HashSet;

import org.jdom.Element;

import ch.epfl.lia.frodo.algorithms.dpop.stochastic.ExpectedUTIL;
import ch.epfl.lia.frodo.solutionSpaces.Addable;
import ch.epfl.lia.frodo.solutionSpaces.AddableReal;
import ch.epfl.lia.frodo.solutionSpaces.BasicUtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.DCOPProblemInterface;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace;
import ch.epfl.lia.frodo.solutionSpaces.UtilitySolutionSpace.ProjOutput;

/** A UTIL propagation phase that maximizes worst-case utility with respect to the random variables
 * @author Thomas Leaute
 * @param <Val> 	the type used for variable values
 */
public class WorstCaseUTIL < Val extends Addable<Val> > extends ExpectedUTIL<Val> {
	
	/** The constructor called in "statistics gatherer" mode
	 * @param problem 		the overall problem
	 * @param parameters 	the description of what statistics should be reported
	 */
	public WorstCaseUTIL(Element parameters, DCOPProblemInterface<Val, AddableReal> problem) {
		super(parameters, problem);
	}

	/** Constructor 
	 * @param problem 		the agent's subproblem
	 * @param parameters 	the parameters for the module
	 */
	public WorstCaseUTIL(DCOPProblemInterface<Val, AddableReal> problem, Element parameters) {
		super.problem = problem;
		problem.setUtilClass(AddableReal.class);
		super.withAnonymVars = true;
		super.method = null;
	}

	/** @see ExpectedUTIL#project(UtilitySolutionSpace, String) */
	@Override 
	protected ProjOutput<Val, AddableReal> project (UtilitySolutionSpace<Val, AddableReal> space, String var) {
		
		// Look up all random variables, and check whether they must all be projected out here
		ArrayList<String> allRandVars = new ArrayList<String> (space.getNumberOfVariables());
		HashSet<String> projHere = this.randVarsProj.get(var);
		boolean projectAll = true;
		for (String randVar : space.getVariables()) {
			if (this.problem.isRandom(randVar)) {
				allRandVars.add(randVar);
				
				if (! projHere.contains(randVar)) 
					projectAll = false;
			}
		}
		
		// Then compute the worst case for the space over all random variables
		UtilitySolutionSpace<Val, AddableReal> worst = space;
		switch (allRandVars.size()) {
		case 0: 
			return space.project(var, this.maximize); // no random variable; we can simply project
		case 1: 
			worst = worst.blindProject(allRandVars.get(0), ! this.maximize);
			break;
		default:
			worst = worst.blindProject(allRandVars.toArray(new String [allRandVars.size()]), ! this.maximize);
		}
		
		// If all random variables must be projected here, just return the projection
		if (projectAll) 
			return worst.project(var, maximize);
		
		// Else, compute the optimal assignments for the projected out variable
		BasicUtilitySolutionSpace< Val, ArrayList<Val> > assignments = worst.project(var, maximize).assignments; /// @todo Implement an argProject() method
		worst = null;

		// Report the corresponding true utilities, as a function of all random variables
		String[] varsOut = new String[] {var};
		return new ProjOutput<Val, AddableReal> (space.compose(varsOut, assignments), varsOut, assignments);
	}
	
	/** @see ExpectedUTIL#sendToParent(String, String, UtilitySolutionSpace) */
	@Override 
	protected void sendToParent (String var, String parent, UtilitySolutionSpace<Val, AddableReal> space) {
		
		// Before sending the UTIL message, we need to project out the random variables
		ArrayList<String> randVars = new ArrayList<String> ();
		for (String randVar : this.randVarsProj.get(var)) 
			if (space.getDomain(randVar) != null) // the space contains this random variable
				randVars.add(randVar);
		if (! randVars.isEmpty()) 
			space = space.blindProject(randVars.toArray(new String [randVars.size()]), ! this.maximize);

		super.sendToParent(var, parent, space);
	}
	
	/** @see ExpectedUTIL#sendOutput(UtilitySolutionSpace, java.lang.String) */
	@Override
	protected void sendOutput(UtilitySolutionSpace<Val, AddableReal> space, String root) {
		
		// First project out all random variables
		ArrayList<String> randVars = new ArrayList<String> (space.getNumberOfVariables());
		for (String randVar : space.getVariables()) 
			if (this.problem.isRandom(randVar)) 
				randVars.add(randVar);
		if (! randVars.isEmpty()) 
			space = space.blindProject(randVars.toArray(new String [randVars.size()]), ! this.maximize);

		super.sendOutput(space, root);
	}
	
}
